--- HazLoot - a Master Looting mod for WoW
-- HazLoot sets two thresholds for loot: hard and soft. Items above the soft threshold (which
--  is the Blizzard master loot threshold), but below the hard threshold are automatically
--  distributed to the raid, randomly and evenly. Items at or above the hard threshold are
--  Phat Lewts: epics that are handled by a configurable looting strategy.
-- HazLoot structures looting strategies in three phases: Auction, Roll, Dust. The last two
--  phases are built-in (going to random rolls and disenchanting). The Auction phase is by
--  default based on a Suicide Kings system. Future versions will include more strategies,
--  including a "Need or Greed" for pugs, DKP, EPGP, Loot Council, etc.
local standalone = false
if LibStub then
  HazLoot = LibStub("AceAddon-3.0"):NewAddon("HazLoot", "AceConsole-3.0", "AceEvent-3.0")
  L = LibStub("AceLocale-3.0"):GetLocale("HazLoot", true)
else -- not in WoW
  HazLoot = {}
  standalone = true
  function HazLoot:Print(message)
    print("[Core] "..message)
  end
  function HazLoot:Debug(message)
    print("[Core][DEBUG] "..message)
  end
  require("HazLootList")
  require("HSK")
  require("Bootstrap")
end

--TODO: Implement timers for bidding process

-- Constants
HazLoot.TRASH     = 0
HazLoot.COMMON    = 1
HazLoot.UNCOMMON  = 2
HazLoot.RARE      = 3
HazLoot.EPIC      = 4
HazLoot.LEGENDARY = 5
HazLoot.ARTIFACT  = 6

HazLoot.LootTypes = {"Common","Uncommon","Rare","Epic","Legendary","Artifact" }
HazLoot.LootTypes[HazLoot.TRASH]     = "Trash"
HazLoot.LootColours = { "ffffff", "1eff00", "0080ff", "b048f8", "ff8000", "ff0000",}
HazLoot.LootColours[HazLoot.TRASH]   = "9d9d9d"

HazLoot.STATES = {}
HazLoot.STATENAMES = {}
local STATE = HazLoot.STATES
local STATES = HazLoot.STATENAMES
STATE.IDLE              = 0 -- Scanning, looting, whatever. Not bidding.
STATE.AUCTION_OPEN      = 1 -- Auction is are open
STATE.LASTCALL          = 2 -- Last call for bids
STATE.ROLL_OPEN         = 3 -- Bids are closed - no valid bids - gone to offspec rolls
STATE.BANKING           = 4 -- Item is bing banked
STATE.DUSTING           = 5 -- Item is bing dusted
STATE.CONFIRMED         = 6 -- Bid winner is being confirmed
STATE.AWARDING          = 7 -- Loot is being awarded
STATES = {"Bid open","Last call","Roll open","Banking","Dusting","Confirming bid winner","Awarding loot!",}
STATES[STATE.IDLE] = "Idle"

local state = STATE.IDLE

local PhatLewt     = {}   -- Table of current distributable loot
local currentItem  = nil  -- loot distro sequence number of loot item being distributed
-- TODO: Move the next five tables to the strategy
local Suicides     = {}   -- record of all "hard" loot suicided for by name (see format below)
local Offspecs     = {}   -- record of all "hard" loot rolled for   ({name={link,link,link,},})
local Banked       = {}   -- record of all "hard" send to the GBank
local Dusted       = {}   -- record of all "hard" disenchanted
HazLoot.PartyMembers = {}   -- Array of current members of the party/raid {name1,name2,...,nameN,}  (max 40)
HazLoot.PartyIndex   = {}   -- Inverted index into PartyMembers {[name1]=1,[name2]=2,...,[nameN]=n} (max 40)
local lootList     = {}   -- Table of random loots by party member: {[quality1]={[name1]=n1,[name2=n2],},[quality2]={...},}
local leastLoot    = {}   -- Table of the least loot received by party members, by quality: {[quality1]=m1,[quality2]=m2,..}
local inRaid       = false
local inParty      = false
local dStrategy      = HSK  -- The distribution strategy (SK currently, DKP, EPGP possible...)
local candidate      = {}   -- The current winning candidate
local candidates     = {}   -- Array of current candidates
local last = {}
last.mainadded  = nil  -- The raiders most recently set as present
last.present    = nil  -- The raiders most queried for presence and SK position
last.dropped    = nil  -- Name of member just dropped. ?? move to HSK/Strategy?
last.addpresent = nil
last.suicided = nil

local GetBootstrapLists = function(self)
  return Bootstrap:GetBootListNames()
end

-- TODO: tidy up the config screen. AGAIN.
local options = {
  name    = "HazLootConfig",
  handler = HazLoot,
  type    = 'group',
  args    = {
-- Visible in the Config UI
    ga = { type = "toggle", name = "Guild Announce", desc = "Enable Epic announcements to the Guild.", get  = "IsGuildAnnounce", set  = "ToggleGuildAnnounce", width = "full", order = 0, },
    banker = { type  = "input", name  = "Banker", desc  = "Set the banker for a raid, defaults to Master Looter", set   = "SetBanker", get   = "GetBanker", order=5, },
    duster = { type  = "input", name  = "Duster", desc  = "Set the designated disenchanter for a raid, defaults to Master Looter", set   = "SetDuster", get   = "GetDuster", order=6, },
    soft = { type = "select", name = "Soft Loot Threshold", desc = "Loot above this quality, but below the hard threshold will be randomly and evenly passed out to party/raid members.", values = HazLoot.LootTypes, get    = "GetSoftThreshold", set    = "SetSoftThreshold", style  = "radio", order  = 11, },
    hard = { type = "select", name = "Hard Loot Threshold", desc = "Loot at or above this quality will be kept for mediated assignment.", values = HazLoot.LootTypes, get    = "GetHardThreshold", set    = "SetHardThreshold", style  = "radio", order  = 12, },
    status = { type = "execute", name = "Report in Chat", desc = "Report who's received what in your chat", func = "StatusReport", order = 20, },
    report = { type = "execute", name = "Report To Raid", desc = "Report who's received what to the raid/party", func = "RaidStatusReport", order = 21, },
    optDStrategy = { type = "select", name = "Distribution Strategy", desc = "The loot distribution strategy you're using, like SK, DKP, EPGP, Need/Greed", values = {["HSK"]="Suicide Kings", ["HNG"]="Need or Greed", }, get =  "GetDStrategy", set =  "SetDStrategy", style = "dropdown", order = 25,},
    optListName = { type = "select", name = "Master List", desc = "The name of the list you're using for the current raid", values = GetBootstrapLists(), get="GetListName", set="SetListName", style = "dropdown", order = 26,},
    usage = { type = "execute", name = "Help", desc = "Print help on the /hl commands to the chat window.", func = "Usage", width = "full", order = 30, },
    debug = { type = "toggle", name = "Debug Messages", desc = "Enable debug messages in your chat.", get  = "IsDebug", set  = "ToggleDebug", width = "full", order = -1, },
-- -- -- -- HIDDEN (Command-line or UI only) -- -- -- --
    state = { type = "execute", name = "Current State", desc = "Report the current state.", func = "ReportCurrentState", hidden = true },
    reset = { type = "execute", name = "Reset Current Loot state", desc = "Reset the current loot state, allowing suicides to re-open", func = "ResetLootState", hidden = true, },
    announce = { type = "execute", name = "Announce next item", desc = "Announce the start of bidding for the next item", func = "AnnouncePhatLewt", hidden = true, },
    next = { type = "execute", name = "OpenNext", desc = "Start suicide bidding for next item", func = "OpenNext", hidden = true, },
    lastcall = { type = "execute", name = "Issue last call", desc = "Last call for bids or rolls on current item", func = "LastCall", hidden = true, },
    offspec = { type = "execute", name = "OpenToRoll", desc = "Call for off-spec rolls for current item", func = "OpenToRoll", hidden = true, },
    addmain = { type  = "input", name  = "Add Main", desc  = "Add a main toon to your auction strategy.", set   = "AddMain", get = "GetLastMainAdded", hidden = true, },
    -- winner = { type  = "input", name  = "Propose a candidate for winner, based on bid strategy", desc  = "Determine the winner from a list of bidders", set   = "SetCandidate", get   = "GetCandidate", hidden = true, },
    confirm = { type  = "input", name  = "Confirm", desc  = "Confirm the winner", usage = "<Suicide LastWinner>", get   = "DoNothing", set   = "Confirm", hidden = true, },
    bank = { type = "execute", name = "Bank", desc = "Send loot to the guild bank trustee", func = "BankCurrentItem", hidden = true, },
    dust = { type = "execute", name = "Disenchant", desc = "Send loot to the designated duster", func = "DustCurrentItem", hidden = true, },
    award = { type = "execute", name = "Award", desc = "Award loot to a winning player", func = "AwardCurrentItem", hidden = true, },
    showpresent = { type = "execute", name = "Show Present List", desc = "Show present suicide list", func = "ShowPresent", hidden = true, },
    hide = { type = "execute", name = "Hide full SK List", desc = "Hide full suicide list", func = "HideFull", hidden = true, },
    show = { type = "execute", name = "Show full SK List", desc = "Show full suicide list", func = "ShowFull", hidden = true, },
    suicide = { type = "input", name = "Suicide", desc = "Suicide member in present suicide list", usage = "/hl suicide <name>", set  = "Suicide", get  = "WhoWasSuicided", hidden = true, },
    drop = { type = "input", name = "Drop", desc = "Drop member in master suicide list", usage = "/hl drop <name>", set  = "Drop", get  = "WhoWasDropped", hidden = true, },
    addpresent = { type  = "input", name  = "AddPresent", desc  = "Mark a member or members as present, regardless of raid membership", usage = "/hl addpresent <name>", get   = "WasSetAsPresent", set   = "SetAsPresent", hidden = true, },
    present = { type  = "input", name  = "Present", desc  = "Find out if a member or members are present, and their position in the list.", usage = "/hl present <name>", get   = "WasPresent", set   = "IsPresent", hidden = true, },
    snapshot = { type = "execute", name = "Snapshot SK List", desc = "Take a snapshot of the SK List (use at the start of a session).", func  = "dStrategy:Snapshot", },
    save = { type = "execute", name = "Save SK List", desc = "Persist the Suicide Kings list.", func  = "SaveState", },
    test = { type = "execute", name = "run an in-game test", desc = "run an in-game test", func = "InGameTest", },
    refresh = { type = "execute", name = "Refresh the list of raiders", desc = "The Blizzard raid UI doesn't always update cleanly. Use this to force a re-scan of the raid.", func = "RefreshMemberList", },
    reboot = { type = "execute", name = "Bootstrap the master list", desc = "Use this to reload your strategy list from file.", func = "RebootList", },
  },
}

-- Shortcuts for chat options
options.args.a    = options.args.award
options.args.an   = options.args.announce
options.args.ap   = options.args.addpresent
options.args.c    = options.args.confirm
options.args.fr   = options.args.fullreset
options.args.l    = options.args.list
options.args.lc   = options.args.lastcall
options.args.n    = options.args.next
options.args.os   = options.args.offspec
options.args.p    = options.args.present
options.args.s    = options.args.showpresent
options.args.w    = options.args.winner

local defaults = {
  profile = {
    debug         = true,
    ga            = true,
    -- useSKList     = true,
    banker        = "player",
    duster        = "player",
    optDStrategy  = "HSK",
    optListname   = "Seraphim",
    soft          = HazLoot.UNCOMMON,
    hard          = HazLoot.EPIC,
  },
}

------- ACE AND ADDON EVENTS ---------

function HazLoot:OnInitialize()
  -- probably vain attempt to improve randomness!
  if (standalone) then
    math.randomseed(os.time()) math.random()
  end
  math.random() math.random()
  -- if not standalone then
    self.db = LibStub("AceDB-3.0"):New("HazLootDB", defaults)
    LibStub("AceConfig-3.0"):RegisterOptionsTable("HazLoot", options)
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("HazLoot", "HazLoot")
    self:RegisterChatCommand("hazloot", "ChatCommand")
    self:RegisterChatCommand("hl", "ChatCommand")
  -- else
    -- self.db = {["profile"]={},}
  -- end
  dStrategy = _G[HazLoot:GetDStrategy()]
  self:Debug("dStrategy is now: "..dStrategy:GetName())
  if not dStrategy:Init(self.db) then
    self:Debug("Failed to init strategy from database, bootstrapping: "..HazLoot:GetListName())
    dStrategy:Bootstrap(HazLoot:GetListName())
  end

  HazLoot:InitQualityTables()
end

function HazLoot:RebootList(info)
  dStrategy:Bootstrap(HazLoot:GetListName())
end

function HazLoot:OnEnable()
  -- Called when the addon is enabled
  HazLoot:ResetMemberList(true) -- load member list, if there is one
  self:Print("I can haz loot!")

  self:RegisterEvent("LOOT_OPENED")
  self:RegisterEvent("LOOT_CLOSED")
  self:RegisterEvent("LOOT_SLOT_CLEARED")
  self:RegisterEvent("PARTY_MEMBERS_CHANGED")
  self:RegisterEvent("RAID_ROSTER_UPDATE")
  self:RegisterEvent("UPDATE_INSTANCE_INFO")
end

function HazLoot:OnDisable()
  self:Debug("Saving SK List")
  HazLoot:SaveState(self.db)
  -- Called when the addon is disabled
end

-------- CORE ----------

function HazLoot:LootThatBag()
  local slots = GetNumLootItems()
  if 0 == slots then CloseLoot() return end -- Ignore empty bags
  PhatLewt = nil-- Reset PhatLewt, because we're about to rescan
  PhatLewt = {} -- Reset PhatLewt, because we're about to rescan

  -- Iterate over the slots in the loot bag
  for i = 1,slots do
    if LootSlotIsCoin(i) then
      LootSlot(i)
    else
      local itemLink     = GetLootSlotLink(i)
      local id           = HazLoot:GetItemId(itemLink)
      local item,lootNow = HLL:GetHazLootItem(id,i)
      if lootNow or item.q < HazLoot:GetSoftThreshold() then -- Just loot it
        LootSlot(i)
      elseif item.q < HazLoot:GetHardThreshold() then
        if "Recipe" == itype or "Quest" == itype then
          table.insert(PhatLewt, item)
        else
          HazLoot:AssignRandomLoot(item.q,item.name,item.link,i,slots)
        end
      else
        table.insert(PhatLewt, item)
      end
    end
  end

  if 0 < (#PhatLewt) then
    HLL:Sort(PhatLewt)
    HazLoot:ShowMeLoot(PhatLewt) -- list it in chat
  end
end

--- Show the loot, as ordered, in the chat window. This will later move to a UI.
function HazLoot:ShowMeLoot(lootTable)
  if not lootTable then
    self:Print("No loot to distribute")
    return false
  end
  if standalone then
    self:Print("Distributing loot in this order: ")
    for i,loot in ipairs(lootTable) do
      self:Print(i..". "..loot.link.." ("..(loot.winner or loot.roles)..")")
    end
  else
    HLUI:ShowLootList(PhatLewt, dStrategy:GetPresentList())
  end
  currentItem = 1
  HLUI:SetHelp(dStrategy:GetHelp(state))
  return true
end

local function InitNextItem()
  for order,loot in pairs(PhatLewt) do
    if not loot.winner then
      currentItem = order
      if state ~= STATE.AUCTION_OPEN then
        HazLoot:Warn("Suicides in raid chat for: "..order..". "..loot.link.." ("..loot.roles..")")
        state = STATE.AUCTION_OPEN
      end
      -- HazLoot:Print(dStrategy:GetNextStates(state))
      HLUI:SetHelp(dStrategy:GetHelp(state))
      break
    end
  end
end

--- Having scanned a corpse for loot, announce it to the raid, in the correct order. Idempotent
function HazLoot:AnnouncePhatLewt()
  if not PhatLewt or 0 == #PhatLewt then
    self:Print("No loot to distribute")
    return false
  end
  HazLoot:Announce("Distributing loot in this order: ")
  for i,loot in ipairs(PhatLewt) do
    HazLoot:Announce(i..". "..loot.link.." ("..(loot.winner and "won by: "..loot.winner or loot.roles)..")")
  end
  self:Debug("/hl |cff1eff00n|rext to start bidding on "..PhatLewt[currentItem].link)
  HLUI:SetHelp(dStrategy:GetHelp(state))
  return true
end

--- Open the next item for bids
-- @returns true if there was an item, and it was opened, false otherwise
function HazLoot:OpenNext()
  if not PhatLewt then
    self:Debug("No loot to distribute.")
    return false
  end
  local usage = dStrategy:IsValidStateChange(state, STATE.AUCTION_OPEN)
  if "string" == type(usage)  then
    self:Print(usage)
    return false
  end

  if state == STATE.IDLE then
    InitNextItem()
  end
  state = STATE.AUCTION_OPEN
  candidate = {}
  HLUI:ShowLootList(PhatLewt, dStrategy:GetPresentList())
  return true
end

--- Issues a last call for an auction
-- @returns true if the call was issued, false if mod was in a bad state
function HazLoot:LastCall()
  local tmp = dStrategy:IsValidStateChange(state, STATE.LASTCALL)
  if "string" == type(tmp)  then
    self:Print(tmp)
    return false
  end
  state = STATE.LASTCALL
  local loot = PhatLewt[currentItem]
  HazLoot:Warn("Last Call for "..currentItem..". "..loot.link.." ("..loot.roles..")")
  self:Print(dStrategy:GetNextStates(state))
  HLUI:SetHelp(dStrategy:GetHelp(state))
  return true
end

--- Open an item up for off-spec rolls
-- @returns true if opened, false if in the wrong state
function HazLoot:OpenToRoll()
  local usage = dStrategy:IsValidStateChange(state, STATE.ROLL_OPEN)
  if "string" == type(usage)  then
    self:Print(usage)
    return false
  end
  state = STATE.ROLL_OPEN
  local loot = PhatLewt[currentItem]
  HazLoot:Warn("Roll offspec for "..currentItem..". "..loot.link.." ("..loot.roles..")")
  self:Debug(" /hl |cff1eff00c|ronfirm  <name> to confirm the winner")
  return true
end

local function GetLootCandidateByName(name)
  local lootCandidate = nil
  for i=1,40 do -- Loop over all possible master loot candidates
    lootCandidate = GetMasterLootCandidate(i)
    if lootCandidate and string.lower(name) == string.lower(lootCandidate) then
      return i,lootCandidate
    end
  end
  return nil
end

--- Defines the target for awarding loot, confirms to the party/raid
-- @returns true if the winner was found, and is a master loot candidate, false if not, or
--          if the command is issued while in the wrong state.
function HazLoot:Confirm(info)
  local candidate = HazLoot:GetCandidate()

  if not candidate.name then
    HLUI:SetHelp("In chat, type /hl |cff1eff00w|rinner <winning roller> to propose a candidate")
    return false
  end

  local usage = dStrategy:IsValidStateChange(state, STATE.CONFIRMED)
  if "string" == type(usage)  then
    self:Print(usage)
    return false
  end

  local index, name = GetLootCandidateByName(candidate.name)
  if not index then
    self:Print(candidate.name.." is not a Master Loot Candidate for "..PhatLewt[currentItem].link)
    return false
  end
  candidate.index    = index -- raidN or partyN
  candidate.name     = name
  candidate.link     = PhatLewt[currentItem].link
  candidate.slot     = PhatLewt[currentItem].slot
  candidate.oldstate = state

  HLUI:SetHelp(dStrategy:GetHelp(state))
  state = STATE.CONFIRMED
  return true
end

function HazLoot:BankCurrentItem(info)
  local usage = dStrategy:IsValidStateChange(state, STATE.BANKING)
  if "string" == type(usage)  then
    self:Print(usage)
    return false
  end
  state = STATE.BANKING
  HazLoot:ClearCandidates()
  HazLoot:AddCandidate(HazLoot:GetBanker())
  return true
end

function HazLoot:DustCurrentItem(info)
  local usage = dStrategy:IsValidStateChange(state, STATE.DUSTING)
  if "string" == type(usage)  then
    self:Print(usage)
    return false
  end
  state = STATE.DUSTING
  HazLoot:ClearCandidates()
  HazLoot:AddCandidate(HazLoot:GetDuster())
  return true
end

function HazLoot:AwardCurrentItem(info)
  HazLoot:Confirm(info)
  local usage = dStrategy:IsValidStateChange(state, STATE.AWARDING)
  if "string" == type(usage)  then
    self:Print(usage)
    return false
  end
  local candidate = HazLoot:GetCandidate()
  if not candidate.name then print("no candidate") return false end
  GiveMasterLoot(candidate.slot,candidate.index)
  return true
end

function HazLoot:LootCleared(slotCleared)
  if not currentItem then return end -- nothing being awarded
  if not PhatLewt then return end    -- nothing worth awarding
  local candidate = HazLoot:GetCandidate()
  for order,loot in pairs(PhatLewt) do
    if slotCleared == loot.slot then
      if     candidate.oldstate == STATE.LASTCALL  then
        Suicides[candidate.name] = Suicides[candidate.name] or {}
        table.insert(Suicides[candidate.name], candidate.link)
        HazLoot:Announce("Grats "..candidate.name.."! "..candidate.link.." ("..dStrategy.auction..")")
        HazLoot:GAnnounce("Grats "..candidate.name.."! "..candidate.link.." ("..dStrategy.auction..")")
      elseif candidate.oldstate == STATE.ROLL_OPEN then
        Offspecs[candidate.name] = Offspecs[candidate.name] or {}
        table.insert(Offspecs[candidate.name], candidate.link)
        HazLoot:Announce("Grats "..candidate.name.."! "..candidate.link.." ("..dStrategy.roll..")")
        HazLoot:GAnnounce("Grats "..candidate.name.."! "..candidate.link.." ("..dStrategy.roll..")")
      elseif candidate.oldstate == STATE.BANKING   then
        table.insert(Banked, candidate.link)
        HazLoot:GAnnounce(candidate.link.." sent to "..HazLoot:GetBanker().." for "..dStrategy.bank)
      elseif candidate.oldstate == STATE.DUSTING   then
        table.insert(Dusted, candidate.link)
        HazLoot:Announce(candidate.link.." sent to "..HazLoot:GetDuster().." for "..dStrategy.dust)
      end

      if candidate.oldstate == STATE.LASTCALL then HazLoot:HandleWinner()
      elseif candidate.oldstate == STATE.ROLL_OPEN then dStrategy:HandleWinner(candidate.name, 0)
      else print("old state was : "..STATES[candidate.oldstate])
      end

      PhatLewt[currentItem].winner = candidate.name
      currentItem = currentItem + 1
      candidate   = {}
      if currentItem > #PhatLewt then
        PhatLewt = nil
        state = STATE.IDLE
        HLUI:CloseLoot()
      else
        state = STATE.IDLE
      end
    end
  end
  HLUI:CloseLoot()
  HLUI:SetHelp(dStrategy:GetHelp(state))
  HLUI:ShowLootList(PhatLewt, dStrategy:GetPresentList())
end

--- Assign loot to a random party member such that everyone gets N before anyone gets N+1 items of a give quality.
-- @returns 
function HazLoot:AssignRandomLoot(itemQuality,itemName,itemLink,slot,slots)
  if standalone then self:Debug("[AssignRandomLoot] "..itemName.." in slot "..slot) end
  -- 1. Pick a random starting point in the list for items of this quality
  -- 2. Cycle through through that list once until:
  -- 3a. EITHER find someone who has leastLoot loots of that quality, and award it
  -- 3b. OR find no nobody with leastLoot lots of that quality, award it to the random start and ++LeastLoot
  -- 4. If the attempted recipient is not a valid master loot candidate, skip them
  -- self:Debug(os.time())
  local tmpRand = math.random(40)
  local looterName
  local failures = {}
  for i = 1,#HazLoot.PartyMembers do
    local randomLooter = (i+tmpRand)%#HazLoot.PartyMembers + 1      -- +1 for LUA 1-based table indexes
    looterName = HazLoot.PartyMembers[randomLooter]
    if nil ~= looterName then
      if (leastLoot[itemQuality] >= lootList[itemQuality][looterName]) then
        if HazLoot:GiveRandomLoot(itemQuality,itemLink,slot,looterName) then
          return {}
        else
          table.insert(failures,looterName)
        end
      end
    end
  end
  -- either everyone has at least LeastLoot items, or everyone who does not is not eligible
  if 0 == #failures then
    looterName = HazLoot.PartyMembers[tmpRand%#HazLoot.PartyMembers + 1]
    leastLoot[itemQuality] = leastLoot[itemQuality] +1
    if HazLoot:GiveRandomLoot(itemQuality,itemLink,slot,looterName) then
      self:Debug("Everyone should have received "..leastLoot[itemQuality].." "..HazLoot.LootTypes[itemQuality].." lewt(s).")
      return {}
    end
  elseif #failures == #HazLoot.PartyMembers then
    self:Debug("Nobody was eligible for "..itemLink..", so I have no idea what to do. I'm probably buggered now.")
  else
    self:Debug("Assigning "..itemLink.." to someone with more than the minimum. Nobody else is eligible.")
    tmpRand = math.random(40) -- get a new seed
    for i = 1,#HazLoot.PartyMembers do
      local randomLooter = (i+tmpRand)%#HazLoot.PartyMembers + 1      -- +1 for LUA 1-based table indexes
      looterName = HazLoot.PartyMembers[randomLooter]
      if looterName and HazLoot:GiveRandomLoot(itemQuality,itemLink,slot,looterName) then
          return true
      end
    end
  end
end

--- Give out loot randomly.
-- @param quality The itemQuality of the loot
-- @param link   The itemLink of the loot
-- @param slot   The loot slot of the corpse being looted
-- @param looter The name of the party member being given the loot
-- @returns true if it was given out successfully, false otherwise
function HazLoot:GiveRandomLoot(quality,link,slot,looter)
  for i=1,40 do -- Might as well loop over all possible slots
    if looter == GetMasterLootCandidate(i) then -- Use GMLC() here, because consistency doesn't matter: loot's been assigned
      GiveMasterLoot(slot,i)
      if nil == lootList[quality][looter] then
        lootList[quality][looter] = 1
      else
        lootList[quality][looter] = lootList[quality][looter] + 1
      end
      HazLoot:Announce(looter.." got "..link.." ("..lootList[quality][looter].." "..HazLoot.LootTypes[quality].."(s) this session).")
      return true
    end
  end
  self:Print("|cffff0000According to Blizzard, "..looter.." was not a valid master loot candidate for|r "..link)
  return false
end

--- Set up the loot tables 
function HazLoot:InitQualityTables()
  for quality = 0,#HazLoot.LootTypes do
    if quality >= HazLoot:GetSoftThreshold() then
      lootList[quality]  = {}
      leastLoot[quality] = 0
    end
  end
end

--- Debug info
function HazLoot:ReportCurrentState(info)
  local tmp = ""
  tmp = tmp.."State: "..STATES[state].."\n"
  if PhatLewt and #PhatLewt > 0 then
    tmp = tmp.."Current loot for distribution: \n"
    for i,lewt in ipairs(PhatLewt) do
      tmp = tmp..i..". "..lewt.link.." ("..lewt.roles..")\n"
    end
  end
  if (HazLoot.last) then
    tmp = tmp.."Most recent activity: \n"
    for what,who in pairs(HazLoot.last) do tmp = "Last "..what.." was: "..who end
  end
  -- self:Debug(tmp)
  return tmp
end

function HazLoot:RaidStatusReport(info)
  HazLoot:GetStatusInfo(true)
end
function HazLoot:StatusReport(info)
  HazLoot:GetStatusInfo(false)
end

function HazLoot:GetStatusInfo(reportToRaid)
  local raider
  local tmpString = "In SK List: "..dStrategy:GetMasterListSize()..", present in raid: "..dStrategy:GetPresentCount()
  if (reportToRaid) then HazLoot:Announce(tmpString) else HazLoot:Debug(tmpString) end
  for i=1,40 do
    raider = HazLoot.PartyMembers[i]
    if (nil ~= raider) then
      tmpString = ""
      for quality = HazLoot:GetSoftThreshold(),HazLoot:GetHardThreshold()-1 do
        tmpString = tmpString..lootList[quality][raider].." "..HazLoot.LootTypes[quality]..", "
      end
      tmpString = tmpString.."to "..raider.." - "

      if Suicides[raider] then
        tmpString = tmpString.." Suicide(s): "
        for i,win in ipairs(Suicides[raider]) do
          tmpString = tmpString..(i>1 and ", " or "")..i.."."..win
        end
      else tmpString = tmpString.." (no suicides)"
      end
      if Offspecs[raider] then
        tmpString = tmpString.." Offspec(s): "
        for i,win in ipairs(Offspecs[raider]) do
          tmpString = tmpString..(i>1 and ", " or "")..i..". "..win
        end
      else tmpString = tmpString.." (no offspec)"
      end

      if (reportToRaid) then HazLoot:Announce(tmpString) else HazLoot:Debug(tmpString) end
    end
  end
  tmpString = ""
  if #Banked > 0 then
    tmpString = tmpString.."Banked: "
    for i,win in ipairs(Banked) do
      tmpString = tmpString..(i>1 and ", " or "")..win
    end
    tmpString = tmpString.."\n"
  end
  if (reportToRaid) then HazLoot:Announce(tmpString) else HazLoot:Debug(tmpString) end
  tmpString = ""
  if #Dusted > 0 then
    tmpString = tmpString.."Dusted: "
    for i,win in ipairs(Dusted) do
      tmpString = tmpString..(i>1 and ", " or "")..win
    end
  end
  if (reportToRaid) then HazLoot:Announce(tmpString) else HazLoot:Debug(tmpString) end
end

function HazLoot:SaveState()
  return dStrategy:SaveState(self.db)
end

function HazLoot:Usage() --TODO: rewrite
  self:Print("Open a target with phat lewts to announce the distributable loots, then use /hazloot or /hl commands:")
  self:Print(" |cffff0000/hl |rn|cffff0000ext|r                       - Open bidding on the next loot item.")
  self:Print(" |cffff0000/hl |rl|cffff0000ast|rc|cffff0000all|r                   - Last call for bids on the current loot item.")
  self:Print(" |cffff0000/hl |ro|cffff0000ff|rs|cffff0000pec|r                    - Open item for off-spec rolls.")
  self:Print(" |cffff0000/hl |rc|cffff0000onfirm|r |cffff0000<winner>|r - Announce <name> as winner for confirmation.")
  self:Print(" |cffff0000/hl |ra|cffff0000ward|r                         - Give current item to most recent confirmed winner.")
  self:Print("Configuration:")
  self:Print(" |cffff0000/hl|r - Bring up the configuration UI.")
  self:Print(" |cffff0000/hl soft|r <quality> - Set the soft threshold to <quality>.")
  self:Print(" |cffff0000/hl hard|r <quality> - Set the hard threshold to <quality>.")
  self:Print("Help:")
  self:Print(" |cffff0000/hl usage|r - Display this message")
end
------- PARTY LIST HANDLING ------

local function CheckRaidOrParty()
  if UnitInRaid("player") then
    inRaid = true
    inParty = false
  elseif UnitInParty("player") then
    inRaid = false
    inParty = true
  end
end

--TODO: look into Ace-Bucket
local function GroupChange()
  if HazLoot.partyListLoaded then
    HazLoot:RefreshMemberList()
  else
    HazLoot:ResetMemberList(true) -- reload
  end
end

function HazLoot:RefreshMemberList()
  local looter = UnitName("player")
  local tmpCurrentParty = {}
  tmpCurrentParty[looter] = 0 -- Don't exclude yourself
  if (nil == dStrategy:MarkAsPresent(looter)) then self:Debug("Could not find "..looter.." in "..dStrategy:GetName().." use /hl add or /hl addalt") end
  -- idempotent Add members still in the raid
  for i = 1,40 do
    looter = UnitName("raid"..i)
    if nil == looter then
      looter = UnitName("party"..i)
    end
    if (nil ~= looter) then
      tmpCurrentParty[looter] = i
      j = HazLoot:AddMemberToParty(looter)
      -- self:Debug("Found party member "..j..": "..looter)
    end
  end
  -- remove anybody who's no longer in the raid
  for i = 1,#HazLoot.PartyMembers do
    if nil == tmpCurrentParty[HazLoot.PartyMembers[i]] then
      HazLoot:UnloadPartyMember(i)
    end
  end
  CheckRaidOrParty()
end

function HazLoot:UnloadPartyMember(index)
  -- TODO: make sure this is the correct behaviour with persistence
  if nil ~= HazLoot.PartyMembers[index] then
    HazLoot.PartyIndex[HazLoot.PartyMembers[index]] = nil
  end
  HazLoot.PartyMembers[index] = nil
  for quality = HazLoot:GetHardThreshold(), HazLoot:GetSoftThreshold(), -1 do
    lootList[quality][index] = nil
  end
end

function HazLoot:FullReset()
  HazLoot:ResetMemberList(true)
  PhatLewt     = {} -- Table of current distributable loot
  currentItem  = nil
  candidate    = {}
  state = STATE.IDLE
end

function HazLoot:ResetLootState()
  PhatLewt     = {} -- Table of current distributable loot
  currentItem  = nil
  candidate    = {}
  state = STATE.IDLE
  self:Print("Close and re-open the corpse to finish resetting")
end

function HazLoot:ResetMemberList(reload)
  for quality = HazLoot:GetHardThreshold(), HazLoot:GetSoftThreshold(), -1 do
    lootList[quality]  = nil
    leastLoot[quality] = 0
  end
  HazLoot.PartyMembers = nil
  HazLoot.PartyIndex   = nil
  HazLoot.PartyIndex   = {}
  HazLoot.PartyMembers = {}
  for i=HazLoot:GetSoftThreshold(), HazLoot:GetHardThreshold() do
    lootList[i] = nil
    lootList[i] = {}
  end
  HazLoot.partyListLoaded = false

  if reload then
    HazLoot.partyListLoaded = HazLoot:LoadPartyMembers()
  end
end

function HazLoot:LoadPartyMembers()
  local looter, j
  for i = 1,40 do
    looter = UnitName(i < 6 and "party"..i or "raid"..i)
    if (nil ~= looter) then
      j = HazLoot:AddMemberToParty(looter)
    end
  end
  -- Don't forget yourself! (Outside of a raid, the loop won't find the player)
  looter = UnitName("player")
  j = HazLoot:AddMemberToParty(looter)
  CheckRaidOrParty()
  return true
end

-- This function should be idempotent
function HazLoot:AddMemberToParty(looter)
  local i = HazLoot.PartyIndex[looter]
  if nil ~= i then return i end -- They're already in the list. Move on.
  -- otherwise, iterate over the max raid table size, looking for the first empty slot
  for i = 1,40 do -- 40 is largest possible raid size, so we can't need more of a table than that
    if nil == HazLoot.PartyMembers[i] then
      HazLoot.PartyMembers[i]    = looter
      HazLoot.PartyIndex[looter] = i
      for quality,name in ipairs(HazLoot.LootTypes) do
        if (quality >= HazLoot:GetSoftThreshold()) then
          lootList[quality][looter] = leastLoot[quality]
        end
      end
      dStrategy:MarkAsPresent(looter)
      if (nil == dStrategy:MarkAsPresent(looter)) then
        self:Debug("Could not find "..looter.." in "..(dStrategy and dStrategy:GetName() or "unknown strategy,").." use /hl add or /hl addalt")
      -- else
        -- self:Debug(looter.." marked as present in "..(dStrategy and dStrategy:GetName() or "unknown strategy,"))
      end
      return i
    elseif looter == HazLoot.PartyMembers[i] then -- this path should now be redundant
      self:Debug("ERROR: "..looter.." is already party member #"..i.." but I should already know that")
      return i
    end
  end
  self:Debug("I couldn't find room for "..looter.." have you got more than 40 in the raid, or am I broken?")
  return -1
end

------- UTILITY FUNCTIONS ------

--- Extract only the itemId from an itemLink using only Lua native functions
-- itemString: "item:itemId:enchantId:jewelId1:jewelId2:jewelId3:jewelId4:suffixId:uniqueId"
-- @params itemLink A WoW itemLink
-- @returns The numeric itemId of the itemLink
function HazLoot:GetItemId(itemLink)
  if not itemLink then return 0 end
  local found, _, itemString = string.find(itemLink, "^|c%x+|H(.+)|h%[.*%]")
  local itemId = string.sub(itemString,6,string.find(itemString,":",7)-1)
  return tonumber(itemId)
end

function HazLoot:Warn(message)
  if not standalone then
    SendChatMessage(message, "RAID_WARNING")
  else
    self:Debug("[RAID_WARNING]"..message)
  end
end

function HazLoot:GAnnounce(message)
  if HazLoot:IsGuildAnnounce() and not standalone then
    SendChatMessage(message, "GUILD")
  else
    self:Debug("[GUILD]"..message)
  end
end

function HazLoot:Announce(message)
  if inRaid then
    SendChatMessage(message, "RAID")
  elseif inParty and not standalone then
    SendChatMessage(message, "PARTY")
  else
    self:Debug("inParty is "..(inParty and "true" or "false")..", while standalone is "..(standalone and "true" or "false"))
    self:Debug(message)
  end
end

function HazLoot:Debug(message)
  if HazLoot:IsDebug() then
      self:Print(message)
  end
end

function HazLoot:QualityMessage(message,quality)
  if (nil == HazLoot.LootColours[quality]) then return message end
  return "|cff"..HazLoot.LootColours[quality]..message.."|r"
end

function HazLoot:ColouredQuality(quality)
  return HazLoot:QualityMessage(HazLoot.LootTypes[quality], quality)
end
function HazLoot:PlainQuality(colouredQuality)
  -- "|cAARRGGBBQuality|r"
  -- "123456789012345678"
  return string.sub(colouredQuality,11,16)
end

-- inelegant, but functional
function HazLoot:Split(input, separator)
  if not input then return {""} end
  local array = {}
  local j = 1
  local k = 0
  local fragment = nil
  while j < string.len(input) do
    k = string.find(input," ",j)
    if k then
      fragment = string.sub(input, j, k - 1)
      j = k + 1
    else
      fragment = string.sub(input, j)
      j = string.len(input) 
    end
    table.insert(array,fragment)
  end
  return array
end
-------------- STRATEGY INTERACTION --------------

--- Marks a member (or members) as present
function HazLoot:SetAsPresent(info, array)
  local goodString = ""
  local badString = ""
  local result = nil
  for i,member in ipairs(HazLoot:Split(array, " ")) do
    result = dStrategy:MarkAsPresent(member)
    if not result then -- False. Not present.
      badString = badString.." "..member
    elseif "string" == type(result) then -- alt
      goodString = goodString == "" and goodString or goodString.."\n"
      goodString = goodString..dStrategy:GetMasterPosition(result)..". "..member.." ("..result..")"
    else -- result should be true.
      goodString = goodString == "" and goodString or goodString.."\n"
      goodString = goodString..dStrategy:GetMasterPosition(member)..". "..member
    end
  end
  if not result then return false end -- result never got set
  last.addpresent = goodString
  if "" ~= badString then self:Debug("Unknown: "..badString) end
  return goodString
end
--- Returns the position name(s) - and main(s) - of the raider most recently marked as present
function HazLoot:WasSetAsPresent(info)
  return last.addpresent
end

--- Determines whether a member (or list of members) is present in the raid.
function HazLoot:IsPresent(info, array)
  local goodString = ""
  local badString = ""
  local result = nil
  for i,member in ipairs(HazLoot:Split(array, " ")) do
    result = dStrategy:IsPresent(member)
    if not result then -- False. Not present.
      badString = badString.." "..member
    elseif "string" == type(result) then -- alt
      goodString = goodString == "" and goodString or goodString.."\n"
      goodString = goodString..dStrategy:GetMasterPosition(member)..". "..member.." ("..result..")"
    else -- result should be true, so member is present
      goodString = goodString == "" and goodString or goodString.."\n"
      goodString = goodString..dStrategy:GetMasterPosition(member)..". "..member
    end
  end
  if "" ~= goodString then
    last.present = goodString
    self:Debug("Present: "..goodString)
    return goodString
  end
  if "" ~= badString then
    self:Debug("Not present: "..badString)
    return nil
  end
end
function HazLoot:WasPresent(info)
  return last.present
end

function HazLoot:GetLastMainAdded(info)
  return last.mainadded
end
--- Checks the strategy for a list of bidders, and sets the winner into a variable.
-- @returns true if a winner was found, false if none of the bidders was eligible.
function HazLoot:AddMain(info, array)
  local tbl = HazLoot:Split(array, " ")
  local name = tbl[1] -- names must be one word.
  if not name then return false end
  local role = ""
  for i,rolefrag in ipairs(tbl) do
    if i > 2 then role = role.." " end
    if i > 1 then role = role..rolefrag end
  end
  local position = dStrategy:AddMember(name, role)
  if position then
    HazLoot:Print(dStrategy:GetMemberAt(position).." was added at position: "..position)
    last.mainadded = dStrategy:GetMemberAt(position)
    return true
  end
  HazLoot:Print("Couldn't add "..name..", with role '"..role.."'")
  return false
end

function HazLoot:SetBanker(info, input)
  local banker
  if not input or "" == input then
    banker = UnitName("target")
  else 
    banker = HazLoot:Split(input, " ")[1]
    banker = UnitName(banker)
  end
  if not standalone and (UnitInRaid(banker) or UnitInParty(banker)) then
    self:Debug("Setting banker to: "..banker)
    self.db.profile.banker = banker
  else
    self:Debug("Defaulting banker to you")
    self.db.profile.banker = "player"
  end
end
function HazLoot:GetBanker(info)
  return UnitName(self.db.profile.banker) or UnitName(defaults.profile.banker)
end

function HazLoot:SetDuster(info, input)
  local duster
  if not input or "" == input then
    duster = UnitName("target")
  else 
    duster = HazLoot:Split(input, " ")[1]
    duster = UnitName(duster)
  end
  if not standalone and (UnitInRaid(duster) or UnitInParty(duster)) then
    self:Debug("Setting duster to: "..duster)
    self.db.profile.duster = duster
  else
    self:Debug("Defaulting duster to you")
    self.db.profile.duster = "player"
  end
end
function HazLoot:GetDuster(info)
  return UnitName(self.db.profile.duster) or UnitName(defaults.profile.duster)
end

function HazLoot:GetCandidate(info)
  return candidate
end

--- Checks the strategy for a list of bidders, and sets the winner into a variable.
-- @returns true if a winner was found, false if none of the bidders was eligible.
--[[
function HazLoot:SetCandidate(info, array)
  candidates = HazLoot:Split(array, " ")
  local winner = dStrategy:GetCandidate(candidates)
  if not winner then
    HazLoot:Print("None of '"..array.."' is eligible, try again")
    return false
  end
  candidate = {}
  candidate.name = winner
  HLUI:ShowConfirmation(dStrategy:GetConfirmation(state),winner,candidates)
  HazLoot:Confirm(info)
  return winner
end
-- ]]--

function HazLoot:AddCandidate(name)
  local usage = dStrategy:IsValidStateChange(state, STATE.CONFIRMED)
  if "string" == type(usage)  then
    self:Print(usage)
    return false
  end
  local added = false
  for i=1,#candidates do
    if candidates[i] == name then added = true end
  end
  if not added then table.insert(candidates, name) end
  print("HL: candidate table has length:"..(#candidates))
  local winner = dStrategy:GetCandidate(candidates)
  if not winner then
    HazLoot:Print("None of '"..array.."' is eligible, try again")
    return false
  else
  end
  candidate = {}
  candidate.name = winner
  HLUI:ShowConfirmation(dStrategy:GetConfirmation(state),winner,candidates)
  -- HazLoot:Confirm(info)
  return winner
end

function HazLoot:ClearCandidates()
  candidates = {}
  candidate  = {}
end

function HazLoot:HideFull(info)
  HLUI:Hide()
end

function HazLoot:ShowFull(info)
  HLUI:ShowPlayerList(dStrategy:GetMasterList(), "Master SK List")
end

function HazLoot:ShowPresent(info)
  HLUI:ShowPlayerList(dStrategy:GetPresentList(), "Present")
end

--- Tell the strategy to accept the winning candidate's bid (default: drop in SK list)
-- If the drop succeeds, the winning candidate value is cleared, but remains untouched
--  if the drop fails.
-- @returns true if the drop succeeded, false if not
function HazLoot:HandleWinner(info)
  local tmpCand = candidate
  if not tmpCand then HazLoot:Debug("/hl winner to set winner") return false end --TODO: correct
  -- HazLoot:Debug("dropping "..tmpCand.name)
  local newPosition = dStrategy:HandleWinner(tmpCand.name, 1) -- 1 for suicide
  if newPosition then
    HazLoot:Debug(tmpCand.name.." is now at "..newPosition)
    return newPosition
  end
  HazLoot:Debug("Failed to drop "..tmpCand.name)
  return false
end

function HazLoot:Suicide(info, toSuicide)
  if not toSuicide then HazLoot:Debug("/hl suicide <name> to suicide a member. Use to fix errors.") return false end --TODO: correct
  local newPosition = dStrategy:Suicide(toSuicide)
  if newPosition then
    HazLoot:Debug(toSuicide.." is now at "..newPosition)
    last.suicided = toSuicide
    return newPosition
  end
  HazLoot:Debug("Failed to suicide "..toSuicide)
  return false
end
function HazLoot:WhoWasSuicided(info)
  return last.suicided or "nobody"
end

function HazLoot:Drop(info, toDrop)
  if not toDrop then HazLoot:Debug("/hl drop <name> to drop a member: DO NOT USE FOR LOOTING!") return false end --TODO: correct
  local newPosition = dStrategy:Drop(toDrop)
  if newPosition then
    HazLoot:Debug(toDrop.." is now at "..newPosition)
    last.dropped = toDrop
    return newPosition
  end
  HazLoot:Debug("Failed to drop "..toDrop)
  return false
end
function HazLoot:WhoWasDropped(info)
  return last.dropped or "nobody"
end

------- GETTERS AND SETTERS FOR OPTIONS ---------

function HazLoot:GetListName()
  return self.db.profile.optListname or defaults.profile.optListName
end
function HazLoot:SetListName(info, newListName)
  self.db.profile.optListname = newListName
  self:Debug("List name is now: "..(newListName or "nil"))
end

function HazLoot:GetDStrategy()
  return self.db.profile.dStrategy or defaults.profile.optDStrategy
end
function HazLoot:SetDStrategy(info, newDStrategy)
  -- if self.db.profile.dStrategy == newDStrategy then
  if dStrategy == newDStrategy then
    self:Debug("Strategy unchanged")
  else
    self:Debug("Changing Strategy to: "..(newDStrategy or "nil"))
    self.db.profile.dStrategy = newDStrategy
    dStrategy = _G[newDStrategy]
    dStrategy:Init(self.db)
  end
end

function HazLoot:IsGuildAnnounce(info)
  return standalone and defaults.profile.ga or self.db.profile.guildannounce
end
function HazLoot:ToggleGuildAnnounce(info, newValue)
  self.db.profile.guildannounce = newValue
  self:Debug("GAnnounce is "..(newValue and "ON" or "OFF"))
end

function HazLoot:IsDebug(info)
  return standalone and defaults.profile.debug or self.db.profile.debug
end
function HazLoot:ToggleDebug(info, newValue)
  self.db.profile.debug = newValue
end

-- TODO: fix these so they show the coloured versions in the UI again, without booting the looter
function HazLoot:GetHardThreshold(info)
  return standalone and defaults.profile.hard or self.db.profile.hard
end
function HazLoot:SetHardThreshold(info,newValue)
  self.db.profile.hard = newValue
end
function HazLoot:GetSoftThreshold(info)
  return (standalone and defaults.profile.soft) or self.db.profile.soft
end
function HazLoot:SetSoftThreshold(info,newValue)
  self.db.profile.soft = newValue
end

function HazLoot:ChatCommand(input)
  if not input or input:trim() == "" then
    InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
  else
    LibStub("AceConfigCmd-3.0").HandleCommand(HazLoot, "hl", "HazLoot", input)
  end
end

--- WOW IN-GAME EVENTS ---

function HazLoot:UPDATE_INSTANCE_INFO()
  -- Check to see if this is a raid instance
  -- Check that you're NOT autolooting, if so, turn off autoloot
  -- Check for a duster - if no duster, warn every 30s (persist duster)
end

function HazLoot:PARTY_MEMBERS_CHANGED()
  GroupChange()
end

function HazLoot:RAID_ROSTER_UPDATE()
  GroupChange()
end

function HazLoot:LOOT_OPENED()
  lootmethod, mlPartyID, mlRaidID = GetLootMethod()
  if "master" == lootmethod and 0 == mlPartyID then
    HazLoot:LootThatBag()
  elseif "master" == lootmethod then
    self:Debug("Not master looter, HazLoot not in effect")
  end
end

function HazLoot:LOOT_CLOSED()
  HLUI:CloseLoot()
end

function HazLoot:LOOT_SLOT_CLEARED(event, slot)
  HazLoot:LootCleared(slot)
end

-------- TESTING STUFF --------

function HazLoot:InGameTest()
  local tmp = ""
  local testTable = {}
  local ids = {42103, 39717,  40074, 40278, 39734} --40256, 40284, 40637,37401,39558,}
  local name, link, quality, ilevel, mlevel, itype, subtype, stack, loc, texture 
  local i = 1
  for _,id in pairs(ids) do
    name, link, q, ilevel, mlevel, itype, subtype, stack, loc, texture = GetItemInfo(id)
    table.insert(testTable, HLL:GetHazLootItem(id,i))
    tmp = tmp..name.." "..link.." loc: "..loc.."\n"
    i = i + 1
  end
  HLL:Sort(testTable)
  HazLoot:ShowMeLoot(testTable) -- list it in chat
  HLUI:ShowLootList(testTable, {[1]="Haz"})
  PhatLewt = testTable
end
